@@ -40,6 +40,7 @@
 #include "qquickaction_p.h"
 #include "qquickaction_p_p.h"
 #include "qquickshortcutcontext_p_p.h"
+#include "qquickdeferredexecute_p_p.h"
 
 #include <QtGui/qstylehints.h>
 #include <QtGui/qguiapplication.h>
@@ -50,10 +51,6 @@
 
 QT_BEGIN_NAMESPACE
 
-// copied from qabstractbutton.cpp
-static const int AUTO_REPEAT_DELAY = 300;
-static const int AUTO_REPEAT_INTERVAL = 100;
-
 /*!
     \qmltype AbstractButton
     \inherits Control
@@ -116,35 +113,29 @@ static const int AUTO_REPEAT_INTERVAL = 100;
     This signal is emitted when the button is interactively double clicked by the user via touch or mouse.
 */
 
-QQuickAbstractButtonPrivate::QQuickAbstractButtonPrivate()
-    : down(false),
-      explicitDown(false),
-      pressed(false),
-      keepPressed(false),
-      checked(false),
-      checkable(false),
-      autoExclusive(false),
-      autoRepeat(false),
-      wasHeld(false),
-      holdTimer(0),
-      delayTimer(0),
-      repeatTimer(0),
-#if QT_CONFIG(shortcut)
-      shortcutId(0),
-#endif
-      pressButtons(Qt::NoButton),
-      indicator(nullptr),
-      group(nullptr),
-      display(QQuickAbstractButton::TextBesideIcon),
-      action(nullptr)
+void QQuickAbstractButtonPrivate::setPressPoint(const QPointF &point)
 {
+    pressPoint = point;
+    setMovePoint(point);
+}
+
+void QQuickAbstractButtonPrivate::setMovePoint(const QPointF &point)
+{
+    Q_Q(QQuickAbstractButton);
+    bool xChange = !qFuzzyCompare(point.x(), movePoint.x());
+    bool yChange = !qFuzzyCompare(point.y(), movePoint.y());
+    movePoint = point;
+    if (xChange)
+        emit q->pressXChanged();
+    if (yChange)
+        emit q->pressYChanged();
 }
 
 void QQuickAbstractButtonPrivate::handlePress(const QPointF &point)
 {
     Q_Q(QQuickAbstractButton);
     QQuickControlPrivate::handlePress(point);
-    pressPoint = point;
+    setPressPoint(point);
     q->setPressed(true);
 
     emit q->pressed();
@@ -161,6 +152,7 @@ void QQuickAbstractButtonPrivate::handleMove(const QPointF &point)
 {
     Q_Q(QQuickAbstractButton);
     QQuickControlPrivate::handleMove(point);
+    setMovePoint(point);
     q->setPressed(keepPressed || q->contains(point));
 
     if (!pressed && autoRepeat)
@@ -174,6 +166,7 @@ void QQuickAbstractButtonPrivate::handleRelease(const QPointF &point)
     Q_Q(QQuickAbstractButton);
     QQuickControlPrivate::handleRelease(point);
     bool wasPressed = pressed;
+    setPressPoint(point);
     q->setPressed(false);
     pressButtons = Qt::NoButton;
 
@@ -236,14 +229,14 @@ void QQuickAbstractButtonPrivate::startRepeatDelay()
 {
     Q_Q(QQuickAbstractButton);
     stopPressRepeat();
-    delayTimer = q->startTimer(AUTO_REPEAT_DELAY);
+    delayTimer = q->startTimer(repeatDelay);
 }
 
 void QQuickAbstractButtonPrivate::startPressRepeat()
 {
     Q_Q(QQuickAbstractButton);
     stopPressRepeat();
-    repeatTimer = q->startTimer(AUTO_REPEAT_INTERVAL);
+    repeatTimer = q->startTimer(repeatInterval);
 }
 
 void QQuickAbstractButtonPrivate::stopPressRepeat()
@@ -283,6 +276,42 @@ void QQuickAbstractButtonPrivate::ungrabShortcut()
 }
 #endif
 
+void QQuickAbstractButtonPrivate::actionTextChange()
+{
+    Q_Q(QQuickAbstractButton);
+    if (explicitText)
+        return;
+
+    q->buttonChange(QQuickAbstractButton::ButtonTextChange);
+}
+
+void QQuickAbstractButtonPrivate::setText(const QString &newText, bool isExplicit)
+{
+    Q_Q(QQuickAbstractButton);
+    const QString oldText = q->text();
+    explicitText = isExplicit;
+    text = newText;
+    if (oldText == q->text())
+        return;
+
+    q->buttonChange(QQuickAbstractButton::ButtonTextChange);
+}
+
+void QQuickAbstractButtonPrivate::updateEffectiveIcon()
+{
+    Q_Q(QQuickAbstractButton);
+    // We store effectiveIcon because we need to be able to tell if the icon has actually changed.
+    // If we only stored our icon and the action's icon, and resolved in the getter, we'd have
+    // no way of knowing what the old value was here. As an added benefit, we only resolve when
+    // something has changed, as opposed to doing it unconditionally in the icon() getter.
+    const QQuickIcon newEffectiveIcon = action ? icon.resolve(action->icon()) : icon;
+    if (newEffectiveIcon == effectiveIcon)
+        return;
+
+    effectiveIcon = newEffectiveIcon;
+    emit q->iconChanged();
+}
+
 void QQuickAbstractButtonPrivate::click()
 {
     Q_Q(QQuickAbstractButton);
@@ -294,7 +323,7 @@ void QQuickAbstractButtonPrivate::trigger()
 {
     Q_Q(QQuickAbstractButton);
     if (action && action->isEnabled())
-        action->trigger(q); // -> click()
+        QQuickActionPrivate::get(action)->trigger(q, false);
     else if (effectiveEnable)
         emit q->clicked();
 }
@@ -308,6 +337,26 @@ void QQuickAbstractButtonPrivate::toggle(bool value)
         emit q->toggled();
 }
 
+static inline QString indicatorName() { return QStringLiteral("indicator"); }
+
+void QQuickAbstractButtonPrivate::cancelIndicator()
+{
+    Q_Q(QQuickAbstractButton);
+    quickCancelDeferred(q, indicatorName());
+}
+
+void QQuickAbstractButtonPrivate::executeIndicator(bool complete)
+{
+    Q_Q(QQuickAbstractButton);
+    if (indicator.wasExecuted())
+        return;
+
+    if (!indicator || complete)
+        quickBeginDeferred(q, indicatorName(), indicator);
+    if (complete)
+        quickCompleteDeferred(q, indicatorName(), indicator);
+}
+
 QQuickAbstractButton *QQuickAbstractButtonPrivate::findCheckedButton() const
 {
     Q_Q(const QQuickAbstractButton);
@@ -396,23 +445,19 @@ QQuickAbstractButton::~QQuickAbstractButton()
 QString QQuickAbstractButton::text() const
 {
     Q_D(const QQuickAbstractButton);
-    return d->text;
+    return d->explicitText || !d->action ? d->text : d->action->text();
 }
 
 void QQuickAbstractButton::setText(const QString &text)
 {
     Q_D(QQuickAbstractButton);
-    if (d->text == text)
-        return;
+    d->setText(text, true);
+}
 
-#if QT_CONFIG(shortcut)
-    setShortcut(QKeySequence::mnemonic(text));
-#endif
-
-    d->text = text;
-    setAccessibleName(d->text);
-    buttonChange(ButtonTextChange);
-    emit textChanged();
+void QQuickAbstractButton::resetText()
+{
+    Q_D(QQuickAbstractButton);
+    d->setText(QString(), false);
 }
 
 /*!
@@ -580,6 +625,17 @@ void QQuickAbstractButton::setAutoExclusive(bool exclusive)
     emit autoExclusiveChanged();
 }
 
+/*!
+    \qmlproperty bool QtQuick.Controls::AbstractButton::autoRepeat
+
+    This property holds whether the button repeats \l pressed(), \l released()
+    and \l clicked() signals while the button is pressed and held down.
+
+    The default value is \c false.
+
+    The initial delay and the repetition interval are defined in milliseconds
+    by \l autoRepeatDelay and \l autoRepeatInterval.
+*/
 bool QQuickAbstractButton::autoRepeat() const
 {
     Q_D(const QQuickAbstractButton);
@@ -594,7 +650,7 @@ void QQuickAbstractButton::setAutoRepeat(bool repeat)
 
     d->stopPressRepeat();
     d->autoRepeat = repeat;
-    buttonChange(ButtonAutoRepeatChange);
+    emit autoRepeatChanged();
 }
 
 /*!
@@ -604,7 +660,9 @@ void QQuickAbstractButton::setAutoRepeat(bool repeat)
 */
 QQuickItem *QQuickAbstractButton::indicator() const
 {
-    Q_D(const QQuickAbstractButton);
+    QQuickAbstractButtonPrivate *d = const_cast<QQuickAbstractButtonPrivate *>(d_func());
+    if (!d->indicator)
+        d->executeIndicator();
     return d->indicator;
 }
 
@@ -614,14 +672,18 @@ void QQuickAbstractButton::setIndicator(QQuickItem *indicator)
     if (d->indicator == indicator)
         return;
 
-    QQuickControlPrivate::destroyDelegate(d->indicator, this);
+    if (!d->indicator.isExecuting())
+        d->cancelIndicator();
+
+    delete d->indicator;
     d->indicator = indicator;
     if (indicator) {
         if (!indicator->parentItem())
             indicator->setParentItem(this);
         indicator->setAcceptedMouseButtons(Qt::LeftButton);
     }
-    emit indicatorChanged();
+    if (!d->indicator.isExecuting())
+        emit indicatorChanged();
 }
 
 /*!
@@ -642,17 +704,14 @@ void QQuickAbstractButton::setIndicator(QQuickItem *indicator)
 QQuickIcon QQuickAbstractButton::icon() const
 {
     Q_D(const QQuickAbstractButton);
-    return d->icon;
+    return d->effectiveIcon;
 }
 
 void QQuickAbstractButton::setIcon(const QQuickIcon &icon)
 {
     Q_D(QQuickAbstractButton);
-    if (d->icon == icon)
-        return;
-
     d->icon = icon;
-    emit iconChanged();
+    d->updateEffectiveIcon();
 }
 
 /*!
@@ -707,12 +766,14 @@ void QQuickAbstractButton::setAction(QQuickAction *action)
     if (d->action == action)
         return;
 
+    const QString oldText = text();
+
     if (QQuickAction *oldAction = d->action.data()) {
         QQuickActionPrivate::get(oldAction)->unregisterItem(this);
         QObjectPrivate::disconnect(oldAction, &QQuickAction::triggered, d, &QQuickAbstractButtonPrivate::click);
+        QObjectPrivate::disconnect(oldAction, &QQuickAction::textChanged, d, &QQuickAbstractButtonPrivate::actionTextChange);
 
-        disconnect(oldAction, &QQuickAction::textChanged, this, &QQuickAbstractButton::setText);
-        disconnect(oldAction, &QQuickAction::iconChanged, this, &QQuickAbstractButton::setIcon);
+        QObjectPrivate::disconnect(oldAction, &QQuickAction::iconChanged, d, &QQuickAbstractButtonPrivate::updateEffectiveIcon);
         disconnect(oldAction, &QQuickAction::checkedChanged, this, &QQuickAbstractButton::setChecked);
         disconnect(oldAction, &QQuickAction::checkableChanged, this, &QQuickAbstractButton::setCheckable);
         disconnect(oldAction, &QQuickAction::enabledChanged, this, &QQuickItem::setEnabled);
@@ -721,45 +782,78 @@ void QQuickAbstractButton::setAction(QQuickAction *action)
     if (action) {
         QQuickActionPrivate::get(action)->registerItem(this);
         QObjectPrivate::connect(action, &QQuickAction::triggered, d, &QQuickAbstractButtonPrivate::click);
+        QObjectPrivate::connect(action, &QQuickAction::textChanged, d, &QQuickAbstractButtonPrivate::actionTextChange);
 
-        connect(action, &QQuickAction::textChanged, this, &QQuickAbstractButton::setText);
-        connect(action, &QQuickAction::iconChanged, this, &QQuickAbstractButton::setIcon);
+        QObjectPrivate::connect(action, &QQuickAction::iconChanged, d, &QQuickAbstractButtonPrivate::updateEffectiveIcon);
         connect(action, &QQuickAction::checkedChanged, this, &QQuickAbstractButton::setChecked);
         connect(action, &QQuickAction::checkableChanged, this, &QQuickAbstractButton::setCheckable);
         connect(action, &QQuickAction::enabledChanged, this, &QQuickItem::setEnabled);
 
-        QQuickIcon actionIcon = action->icon();
-
-        QString name = actionIcon.name();
-        if (!name.isEmpty())
-            d->icon.setName(name);
-
-        QUrl source = actionIcon.source();
-        if (!source.isEmpty())
-            d->icon.setSource(source);
-
-        int width = actionIcon.width();
-        if (width > 0)
-            d->icon.setWidth(width);
-
-        int height = actionIcon.height();
-        if (height)
-            d->icon.setHeight(height);
-
-        QColor color = actionIcon.color();
-        if (color != Qt::transparent)
-            d->icon.setColor(color);
-
-        setText(action->text());
         setChecked(action->isChecked());
         setCheckable(action->isCheckable());
         setEnabled(action->isEnabled());
     }
 
     d->action = action;
+
+    if (oldText != text())
+        buttonChange(ButtonTextChange);
+
+    d->updateEffectiveIcon();
+
     emit actionChanged();
 }
 
+/*!
+    \since QtQuick.Controls 2.4 (Qt 5.11)
+    \qmlproperty int QtQuick.Controls::AbstractButton::autoRepeatDelay
+
+    This property holds the initial delay of auto-repetition in milliseconds.
+    The default value is \c 300 ms.
+
+    \sa autoRepeat, autoRepeatInterval
+*/
+int QQuickAbstractButton::autoRepeatDelay() const
+{
+    Q_D(const QQuickAbstractButton);
+    return d->repeatDelay;
+}
+
+void QQuickAbstractButton::setAutoRepeatDelay(int delay)
+{
+    Q_D(QQuickAbstractButton);
+    if (d->repeatDelay == delay)
+        return;
+
+    d->repeatDelay = delay;
+    emit autoRepeatDelayChanged();
+}
+
+/*!
+    \since QtQuick.Controls 2.4 (Qt 5.11)
+    \qmlproperty int QtQuick.Controls::AbstractButton::autoRepeatInterval
+
+    This property holds the interval of auto-repetition in milliseconds.
+    The default value is \c 100 ms.
+
+    \sa autoRepeat, autoRepeatDelay
+*/
+int QQuickAbstractButton::autoRepeatInterval() const
+{
+    Q_D(const QQuickAbstractButton);
+    return d->repeatInterval;
+}
+
+void QQuickAbstractButton::setAutoRepeatInterval(int interval)
+{
+    Q_D(QQuickAbstractButton);
+    if (d->repeatInterval == interval)
+        return;
+
+    d->repeatInterval = interval;
+    emit autoRepeatIntervalChanged();
+}
+
 #if QT_CONFIG(shortcut)
 QKeySequence QQuickAbstractButton::shortcut() const
 {
@@ -780,6 +874,40 @@ void QQuickAbstractButton::setShortcut(const QKeySequence &shortcut)
 }
 #endif
 
+/*!
+    \readonly
+    \since QtQuick.Controls 2.4 (Qt 5.11)
+    \qmlproperty real QtQuick.Controls::AbstractButton::pressX
+
+    This property holds the x-coordinate of the last press.
+
+    \note The value is updated on touch moves, but left intact after touch release.
+
+    \sa pressY
+*/
+qreal QQuickAbstractButton::pressX() const
+{
+    Q_D(const QQuickAbstractButton);
+    return d->movePoint.x();
+}
+
+/*!
+    \readonly
+    \since QtQuick.Controls 2.4 (Qt 5.11)
+    \qmlproperty real QtQuick.Controls::AbstractButton::pressY
+
+    This property holds the y-coordinate of the last press.
+
+    \note The value is updated on touch moves, but left intact after touch release.
+
+    \sa pressX
+*/
+qreal QQuickAbstractButton::pressY() const
+{
+    Q_D(const QQuickAbstractButton);
+    return d->movePoint.y();
+}
+
 /*!
     \qmlmethod void QtQuick.Controls::AbstractButton::toggle()
 
@@ -791,6 +919,13 @@ void QQuickAbstractButton::toggle()
     setChecked(!d->checked);
 }
 
+void QQuickAbstractButton::componentComplete()
+{
+    Q_D(QQuickAbstractButton);
+    d->executeIndicator(true);
+    QQuickControl::componentComplete();
+}
+
 bool QQuickAbstractButton::event(QEvent *event)
 {
     Q_D(QQuickAbstractButton);
@@ -819,7 +954,7 @@ void QQuickAbstractButton::keyPressEvent(QKeyEvent *event)
     Q_D(QQuickAbstractButton);
     QQuickControl::keyPressEvent(event);
     if (event->key() == Qt::Key_Space) {
-        d->pressPoint = QPoint(qRound(width() / 2), qRound(height() / 2));
+        d->setPressPoint(QPoint(qRound(width() / 2), qRound(height() / 2)));
         setPressed(true);
 
         if (d->autoRepeat)
@@ -902,6 +1037,15 @@ void QQuickAbstractButton::buttonChange(ButtonChange change)
                 button->setChecked(false);
         }
         break;
+    case ButtonTextChange: {
+        const QString txt = text();
+        setAccessibleName(txt);
+#if QT_CONFIG(shortcut)
+        setShortcut(QKeySequence::mnemonic(txt));
+#endif
+        emit textChanged();
+        break;
+    }
     default:
         break;
     }
@@ -921,7 +1065,7 @@ void QQuickAbstractButton::accessibilityActiveChanged(bool active)
 
     Q_D(QQuickAbstractButton);
     if (active) {
-        setAccessibleName(d->text);
+        setAccessibleName(text());
         setAccessibleProperty("pressed", d->pressed);
         setAccessibleProperty("checked", d->checked);
         setAccessibleProperty("checkable", d->checkable);
